/*
  Copyright 2009 Jason Wells 

  iface.c: input and output routines to interface with Simple DirectMedia Layer.
  All calls to SDL are encapsulated in this file 

  This file is a part of Tic Tac Toe.

  Tic Tac Toe is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  any later version.

  Tic Tac Toe is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Tic Tac Toe.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "SDL/SDL.h"
#include "SDL/SDL_image.h"
#include "SDL/SDL_ttf.h"

#include "error.h"
#include "iface.h"

extern char board[];

static SDL_Surface *screen = NULL;
static TTF_Font *font = NULL;
static SDL_Surface *xs = NULL;
static SDL_Surface *os = NULL;

/* apply: utility function to lay a source surface onto a destination surface at the */
/*        offset (x, y). It only wraps the call to SDL_BlitSurface. */
static void
apply (int x, int y, SDL_Surface *source, SDL_Surface *destination)  
{
  SDL_Rect offset;

  offset.x = x;
  offset.y = y;

  SDL_BlitSurface (source, NULL, destination, &offset);
}

/* loadimg: utility function to open and optimize an image file. This function accepts */
/*          a filename as a string and returns a pointer to the image as an SDL_Surface. */
/*          This image is allocated here and should be freed after use. */
static SDL_Surface *
loadimg (const char *fn)
{
  SDL_Surface *img = NULL;
  SDL_Surface *opt = NULL;
  
  img = IMG_Load (fn);

  if (img != NULL)
    {
      opt = SDL_DisplayFormat (img);
      SDL_FreeSurface (img);
    }
  else
    {
      eprintf ("%s:%d: cannot load image %s: %s\n", __FILE__, __LINE__, fn, SDL_GetError ());
      return NULL;
    }
    
  return opt;
}

/* isinside: test whether the point (x, y) lies inside or on the rectangle r */
static bool
isinside (SDL_Rect *r, int x, int y)
{
  return ((x >= r->x && x <= (r->x + r->w))
	  && (y >= r->y && y <= (r->y + r->h)));
}

/* quitiface: clean up resources in use, called via atexit(3) in initiface() */
static void
quitiface (void)
{
  SDL_FreeSurface (xs);
  SDL_FreeSurface (os);
  TTF_CloseFont (font);
}

/* initiface: set up the SDL environment, window, load images and fonts and set up  */
/*            program cleanup via atexit(3). */
bool
initiface (void)
{
  /* get the ball rolling with SDL */
  if (SDL_Init (SDL_INIT_VIDEO))
    {
      eprintf ("%s:%d: cannot initialize SDL: %s\n", __FILE__, __LINE__, SDL_GetError ());
      return false;
    }

  atexit (SDL_Quit);

  /* set up a window and give it a title */
  screen = SDL_SetVideoMode (640, 480, 32, SDL_SWSURFACE);
  if (screen == NULL)
    {
      eprintf ("%s:%d: failed to set video mode: %s\n", __FILE__, __LINE__, SDL_GetError ());
      return false;
    }
  
  SDL_WM_SetCaption ("Tic-tac-toe", NULL);

  /* set up background */
  SDL_Surface *bg = loadimg ("paper.png");
  if (bg == NULL)
    {
      return false;
    }

  apply (0, 0, bg, screen);
  if (SDL_Flip (screen) == -1)
    {
      return false;
    }

  /* set up fonts */
  if (TTF_Init () == -1)
    {
      eprintf ("%s:%d: cannot set up fonts: %s\n", __FILE__, __LINE__, SDL_GetError ());
      return false;
    }

  atexit (TTF_Quit);

  font = TTF_OpenFont ("nihihand.ttf", 96);
  if (font == NULL)
    {
      eprintf ("%s:%d: cannot open handsean.ttf: %s\n", __FILE__, __LINE__, SDL_GetError ());
      return false;
    }

  /* create an X and O image */
  const SDL_Color color = {0, 0, 0};
  
  xs = TTF_RenderText_Solid (font, "X", color);
  os = TTF_RenderText_Solid (font, "O", color);
  if (xs == NULL || os == NULL)
    {
      eprintf ("%s:%d: cannot render X and O surfaces: %s\n",
	       __FILE__, __LINE__, SDL_GetError ());
    }

  atexit (quitiface);
  
  return true;
}

/* move: receive coordinates of a mouse click and return an integer from [0..9] that */
/*       corresponds to a location on the tic-tac-toe board. */
int
move (void)
{
  SDL_Event e;
  SDL_Rect box = {0, 0, 200, 150};

  bool done = false;
  while (!done)
    {
      SDL_WaitEvent (&e);

      /* if the player clicked a button... */
      if (e.type == SDL_MOUSEBUTTONDOWN)
	{
	  done = true;

	  /* convert from pixels into a particular box location */
	  for (int x = 0; x < 3; x++)
	    {
	      box.x = x * 200;
	      for (int y = 0; y < 3; y++)
		{
		  box.y = y * 150;
		  if (isinside (&box, e.button.x, e.button.y))
		    {
		      return (y * 3) + x;
		    }
		}
	    }
	}
      else
	/* if the player quit the game, then quit! */
	if (e.type == SDL_QUIT)
	  {
	    exit (EXIT_SUCCESS);
	  }
    }

  /* catch errors: we should never get here */
  eprintf ("%s:%d: event handling fail: %d", __FILE__, __LINE__, e.type);
  exit (EXIT_FAILURE);
}

/* show: parse the board[] array and display X's and O's on the window */
void
show (void)
{
  for (int i = 0; i < 3; i++)
    {
      for (int j = 0; j < 3; j++)
	{
	  /* grab a character */
	  char c = board[(i * 3) + j];

	  /* calculate a pixel location for it */
	  int n = 175 * j + 100;
	  int m = 150 * i + 35;

	  /* draw the appropriate character */
	  if (c == 'X')
	    apply (n, m, xs, screen);
	  else
	    if (c == 'O')
	      apply (n, m, os, screen);
	}
    }

  if (SDL_Flip (screen) == -1)
    {
      eprintf ("%s:%d: cannot flip screen: %s\n", __FILE__, __LINE__, SDL_GetError ());
    }
}

/* winalert: display a message for the winner. the winning player is named in the  */
/*           parameter c */
void
winalert (char c)
{
  char s[] = "_ wins!";
  SDL_Surface *message = NULL;
  SDL_Color color = {255, 0, 0};

  /* generate, allocate and render message text  */
  s[0] = c;
  message = TTF_RenderText_Solid (font, s, color);
  if (message == NULL)
    {
      eprintf ("%s:%d: cannot render message: %s: %s\n", __FILE__, __LINE__,
	       s, SDL_GetError ());
    }

  /* center to text in the screen */
  apply ((640 - message->w) / 2, (480 - message->h) / 2, message, screen);
  if (SDL_Flip (screen) == -1)
    {
      eprintf ("%s:%d: cannot flip screen: %s\n", __FILE__, __LINE__, SDL_GetError ());
    }

  SDL_Delay (2500);
  SDL_FreeSurface (message);
}
